! WHIZARD 1.23 Dec 10 2002
! 
! (C) 1999-2002 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module mappings

  use kinds, only: default !NODEP!
  use limits, only: CT_MAPPING_DEFAULT_SCALE
  use diagnostics, only: msg_fatal
  use constants, only: pi
  use parameters, only: parameter_set !NODEP!
  use parameters, only: write !NODEP!
  use particles
  use decay_trees, only: TC
  use divisions !NODEP!

  implicit none
  private


  type, public :: mapping
     private
     integer :: mode
     integer :: particle_code
     real(kind=default) :: mass, width
  end type mapping


  public :: write
  public :: set
  public :: operator(==)
  public :: apply, get

  integer, parameter, public :: &
       & NO_MAPPING = 0, S_CHANNEL = 1, T_CHANNEL =  2, U_CHANNEL = -2, &
       & SPLITTING = 4
  integer, parameter, public :: ON_SHELL = -1
  type(mapping), parameter, public :: &
       & mapping_null = &
       & mapping(NO_MAPPING, 0, 0._default, 0._default)

  interface write
     module procedure mapping_write_unit
  end interface
  interface set
     module procedure mapping_set_code
     module procedure mapping_set_mass
  end interface
  interface operator(==)
     module procedure mapping_equal
  end interface
  interface apply
     module procedure mapping_apply_msq
     module procedure mapping_apply_ct
  end interface
  interface get
     module procedure mapping_get_msq
     module procedure mapping_get_ct
  end interface

contains

  subroutine mapping_write_unit(u, map, k)
    integer, intent(in) :: u
    type(mapping), intent(in) :: map
    integer(kind=TC), intent(in) :: k
    character(len=9) :: str
    select case(map%mode)
    case(S_CHANNEL); str = "s-channel"
    case(SPLITTING); str = "splitting"
    case(T_CHANNEL); str = "t-channel"
    case(U_CHANNEL); str = "u-channel"
    case(ON_SHELL);  str = "on-shell"
    end select
    if (map%mode /= NO_MAPPING) &
         & write(u, '(1x,A,I4,A)') &
         &   'Branch #', k, ":  Mapping (" // str // ") for particle " // &
         &   particle_name(map%particle_code)
  end subroutine mapping_write_unit

  subroutine mapping_set_code (map, mode, code)
    type(mapping), intent(inout) :: map
    integer, intent(in) :: mode, code
    map%mode = mode
    map%particle_code = code
  end subroutine mapping_set_code

  subroutine mapping_set_mass (map, par, splitting_scale, threshold_mass, k)
    type(mapping), intent(inout) :: map
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: splitting_scale, threshold_mass
    integer(kind=TC), intent(in), optional :: k
    if (map%mode /= NO_MAPPING) then
       map%mass  = particle_mass (map%particle_code, par)
       if (map%mass < threshold_mass)  map%mass = 0
       if (map%mode == S_CHANNEL) then
          if (map%mass /= 0) then
             map%width = particle_width (map%particle_code, par)
             if (map%width <= 0) then
                call write(6, par)
                if (present(k)) call write (6, map, k)
                call msg_fatal &
                     & (" S-channel resonance must have positive width")
             end if
          else
             map%mass = splitting_scale
             map%width = 0
             map%mode = SPLITTING
          end if
       end if
    end if
  end subroutine mapping_set_mass

  function mapping_equal (m1, m2) result (equal)
    type(mapping), intent(in) :: m1, m2
    logical :: equal
    if (m1%mode == m2%mode) then
       select case (m1%mode)
       case (NO_MAPPING)
          equal = .true.
       case (S_CHANNEL)
          equal = (m1%mass == m2%mass) .and. (m1%width == m2%width)
       case default
          equal = (m1%mass == m2%mass)
       end select
    else
       equal = .false.
    end if
  end function mapping_equal

  subroutine mapping_apply_msq &
       & (map, r, s, msq, msq_min, msq_max, factor)
    type(mapping), intent(in) :: map
    real(kind=default), intent(in) :: r, s, msq_min, msq_max
    real(kind=default), intent(out) :: msq, factor
    real(kind=default) :: rmin, rmax, rr, msq0, msq1, tmp, x
    integer :: mode
    x = r
    factor = 1
    if (s==0)  call msg_fatal (" Applying msq mapping for zero energy")
    mode = map%mode
    msq0 = map%mass**2
    select case(mode)
    case(S_CHANNEL)
       if (msq0 < msq_min .or. msq0 > msq_max) mode = NO_MAPPING
    case(ON_SHELL)
!       if (msq0 < msq_min) call msg_error (" Mass too low for on-shell decay")
    end select
    select case(mode)
    case(NO_MAPPING, T_CHANNEL, U_CHANNEL)
       msq = (1-x) * msq_min + x * msq_max
       factor = factor * (msq_max - msq_min)/s
    case(S_CHANNEL)
       rmin = atan((msq_min-msq0)/(map%mass*map%width))/pi
       rmax = atan((msq_max-msq0)/(map%mass*map%width))/pi
       rr = (1-x)*rmin + x*rmax
       if (rr>-0.5_default .and. rr<.5_default) then
          tmp = tan(rr*pi)
          msq = map%mass * (map%mass + map%width * tmp)
          factor = factor &
               & * (rmax-rmin) * pi * (1 + tmp**2) * (map%mass * map%width)/s 
       else
          msq = 0
          factor = 0
       end if
    case(SPLITTING)
       tmp = log ((msq_max - msq_min)/msq0 + 1)
       msq1 = msq0 * exp (x*tmp)
       msq = msq1 - msq0 + msq_min 
       factor = factor * msq1 * tmp / s
    case(ON_SHELL)
       msq = msq0
    case default
       call msg_fatal ( " Attempt to apply undefined msq mapping")
    end select
  end subroutine mapping_apply_msq

  subroutine mapping_get_msq &
       & (map, r, s, msq, msq_min, msq_max, factor)
    type(mapping), intent(in) :: map
    real(kind=default), intent(out) :: r
    real(kind=default), intent(in) :: s, msq, msq_min, msq_max
    real(kind=default), intent(out) :: factor
    real(kind=default) :: msq0, msq1, tmp, rmin, rmax
    integer :: mode
    factor = 1
    if (s==0) call msg_fatal (" Applying inverse msq mapping for zero energy")
    mode = map%mode
    msq0 = map%mass**2
    select case(mode)
    case(S_CHANNEL)
       if (msq0 < msq_min .or. msq0 > msq_max)  mode = NO_MAPPING
    end select
    select case(mode)
    case(NO_MAPPING, T_CHANNEL, U_CHANNEL)
       if (msq_max/=msq_min) then
          r = (msq-msq_min) / (msq_max-msq_min)
       else
          r = 0
       end if
       factor = factor * (msq_max - msq_min)/s
    case(S_CHANNEL)
       rmin = atan((msq_min-msq0)/(map%mass*map%width))/pi
       rmax = atan((msq_max-msq0)/(map%mass*map%width))/pi
       tmp = (msq - msq0) / (map%mass * map%width)
       r = (atan(tmp)/pi - rmin) / (rmax-rmin)
       factor = factor &
            & * (rmax-rmin) * pi * (1 + tmp**2) * (map%mass * map%width) / s
    case(SPLITTING)
       tmp = log ((msq_max - msq_min)/msq0 + 1)
       msq1 = msq - msq_min + msq0
       r = log (msq1/msq0) / tmp
       factor = factor * msq1 * tmp / s
    case(ON_SHELL)
!       stop 'Attempt to apply multi-channel to an on-shell particle'
    case default
       call msg_fatal (" Attempt to apply undefined (inverse) msq mapping")
    end select
  end subroutine mapping_get_msq

  subroutine mapping_apply_ct(map, prt, decay_p, k1, k2, sqrts, &
       &                      r, ct, st, factor)
    type(mapping), intent(in) :: map
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), dimension(:), intent(in) :: decay_p
    integer(kind=TC), intent(in) :: k1,k2
    real(kind=default), intent(in) :: sqrts, r
    real(kind=default), intent(out) :: ct, st, factor
    real(kind=default) :: x, tmp, E, p, m, eps, msq_sum
    integer(kind=TC) :: k
    factor = 1
    x = r
    select case(map%mode)
    case(NO_MAPPING, S_CHANNEL, SPLITTING, ON_SHELL)
       tmp = 2*(1-x)
       ct = 1 - tmp
       st = sqrt(tmp*(2-tmp))
    case(T_CHANNEL, U_CHANNEL)
       select case(map%mode)
       case(T_CHANNEL); k=k2       ! used for mass and energy
       case(U_CHANNEL); k=k1
       end select
       p = decay_p(k)
       m = particle_mass(prt(k))
       E = sqrt(p**2 + m**2)
       if (E<=0) then
          ct = 1;  st = 0;  factor = 0
       else
          msq_sum = map%mass**2 + m**2
          if (msq_sum == 0)  msq_sum = CT_MAPPING_DEFAULT_SCALE**2
          eps = min(2 * msq_sum / (sqrts * E), 0.4_default)
          tmp = eps**x
          ct = (1 - 2*tmp) / (1 - 2*eps)
          if (ct>=-1 .and. ct<=1) then
             st = sqrt(1 - ct**2)
             factor = factor * (- Log(eps) * tmp / (1 - 2*eps))
          else
             ct = 1;  st = 0;  factor = 0
          end if
       end if
       if (map%mode==U_CHANNEL) ct = -ct  ! backward singularity
    case default
       call msg_fatal (" Attempt to apply undefined ct mapping")
    end select
  end subroutine mapping_apply_ct

  subroutine mapping_get_ct(map, prt, decay_p, k1, k2, sqrts, &
       &                    r, ct, factor)
    type(mapping), intent(in) :: map
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), dimension(:), intent(in) :: decay_p
    integer(kind=TC), intent(in) :: k1,k2
    real(kind=default), intent(in) :: sqrts
    real(kind=default), intent(out) :: r, factor
    real(kind=default), intent(in) :: ct
    real(kind=default) :: E, p, m, eps, leps, tmp, msq_sum
    integer(kind=TC) :: k
    factor = 1
    select case(map%mode)
    case(NO_MAPPING, S_CHANNEL, SPLITTING, ON_SHELL)
       r = (ct+1)/2
    case(T_CHANNEL, U_CHANNEL)
       select case(map%mode)
       case(T_CHANNEL); k=k2
       case(U_CHANNEL); k=k1
       end select
       p = decay_p(k)
       m = particle_mass(prt(k))
       E = sqrt(p**2 + m**2)
       if (E<=0) then
          r = 0;  factor = 0
       else
          msq_sum = map%mass**2 + m**2
          if (msq_sum == 0)  msq_sum = CT_MAPPING_DEFAULT_SCALE**2
          eps = min(2 * msq_sum / (sqrts * E), 0.4_default)
          leps = Log(eps)
          select case(map%mode)
          case(T_CHANNEL);  tmp = (1-(1-2*eps)*ct)/2  ! forward singularity
          case(U_CHANNEL);  tmp = (1+(1-2*eps)*ct)/2  ! backward
          end select
          r = Log(tmp) / leps
          factor = factor * (- leps * tmp / (1-2*eps))
       end if
    case default
       call msg_fatal (" Attempt to apply undefined (inverse) ct mapping")
    end select
  end subroutine mapping_get_ct


end module mappings
